# -*- coding: utf-8 -*-
"""
Created on Mon Oct  7 10:03:02 2019

@author: icruicks
"""

import networkx as nx, numpy as np, pandas as pd
from sklearn.cluster import SpectralClustering
from sknetwork.clustering import Louvain, modularity
from scipy.spatial.distance import jaccard
from sklearn.metrics import normalized_mutual_info_score as NMI
from sklearn.base import ClusterMixin


class GP_MGLA(ClusterMixin):
    """GP-MGLA technique for weighted cluster ensembling
    
    GP-MGLA: turn the objects-by-clusterings into a cluster association matrix (CA)
    that is specially weighted by the clustering qualities relative to the average
    clustering
    
    Parameters
    ----------
    base_clusters : list of array_like or pandas dataframe
        List of cluster labels where each set of labels must be in the same order
        and have shape (n_objects,). Note: standard output from sci-kit learn 
        clustering algorithms will return this format. If pandas dataframe, 
        should be shape (n_objects, n_clusters)
    metaclustering_alg : string
        string to denote how to cluster the object-to-object cluster association
        graph. Options are: 'louvain','spectral','METIS'. Default is 'louvain'
    n_clusters : int
        Number of clusters to find. Used only for spectral and METIS. Default
        is None
    beta : float
        Controls impact of normalized crowd agreement index. Default is 2
    alpha : float
        Controls impact of weighting the cluster association matrix. Default
        is 0.5
    
    Returns
    -------
    z : ndarray
        cluster labels of shape (n,).
    """
    
    def __init__(self, metaclustering_alg = 'louvain', n_clusters = None,
                 beta=2, alpha=0.5):
        
        self.metaclustering_alg = metaclustering_alg
        self.n_clusters = n_clusters
        self.beta = beta
        self.alpha = alpha
    
    def fit_predict(self, base_clusters):
        '''compute the normalized crowd agreement index (NCAI)'''
        NCAI = []
        for cluster in base_clusters:
            CAI = 1/(len(base_clusters)-1) * np.sum([NMI(cluster, alt_cluster) for alt_cluster in base_clusters if alt_cluster is not cluster])
            NCAI.append(CAI*np.ones(np.unique(cluster).size))
        
        NCAI = np.concatenate(NCAI)
        NCAI = NCAI / np.max(NCAI)
        
        '''Transform the list or dataframe of clustering labels into a 
        binary association matrix'''
        ba_matrices =[]
        
        if isinstance(base_clusters, pd.DataFrame):
            base_clusters = [base_clusters.loc[:,i].to_numpy(dtype=int) for i in base_clusters.columns]
        else:
            base_clusters = [clustering.astype(np.int64) for clustering in base_clusters]
        
        for idx in range(len(base_clusters)):
            base_cluster = base_clusters[idx]
            ba_matrix = np.zeros((base_cluster.size, base_cluster.max()+1))
            ba_matrix[np.arange(base_cluster.size), base_cluster] = 1
            ba_matrices.append(ba_matrix)
            
        ba_matrix = np.concatenate(ba_matrices, axis=1)
        
        '''compute the SACT similarity measure between the clusters of the 
        base clusterings'''
        SACT_coeff = np.zeros((ba_matrix.shape[1], ba_matrix.shape[1]))
        
        for i in range(ba_matrix.shape[1]):
            for j in range(i+1, ba_matrix.shape[1]):
                if np.asarray(ba_matrix[:,i]* ba_matrix[:,j] != 0).nonzero()[0].size ==0:
                    for k in range(ba_matrix.shape[1]):
                        SACT = NCAI[k]**self.beta * np.minimum(1-jaccard(ba_matrix[:,i], ba_matrix[:,k]), 1-jaccard(ba_matrix[:,j], ba_matrix[:,k]))
                        SACT_coeff[i,j] +=SACT
                        SACT_coeff[j,i] +=SACT
                    
        SACT_sim = SACT_coeff / np.max(SACT_coeff)
        np.fill_diagonal(SACT_sim, 1)
        
        ''' Create a full object by cluster graph and cluster to cluster graph
        and combine them into one graph for final paritioning'''
        n_entities = ba_matrix.shape[0]
        n_clusters = ba_matrix.shape[1]
        final_graph = np.zeros((n_entities+n_clusters, n_entities+n_clusters))
        final_graph[n_entities:n_entities+n_clusters, 0:n_entities] = self.alpha *np.transpose(ba_matrix*NCAI**self.beta)
        final_graph[0:n_entities, n_entities:n_entities+n_clusters] = self.alpha *(ba_matrix*NCAI**self.beta)
        final_graph[n_entities:n_entities+n_clusters, n_entities:n_entities+n_clusters] = SACT_sim
        
        final_clusterer = graph_clustering_alg()
        self.graph = nx.from_numpy_array(final_graph)
        final_clusters = final_clusterer.cluster(final_graph, self.n_clusters, self.metaclustering_alg)
        self.entity_clusters = final_clusters[0:n_entities]
        self.cluster_clusters = final_clusters[n_entities-1:]
        return self.entity_clusters
        
        
class CSPA(ClusterMixin):
    """CSPA+ is the mainstream technique for cluster ensembling
    
    CSPA: turn the objects-by-clusterings into a cluster association matrix (CA)
    From the original CSPA algorithm, several modifications have been added including
    thresholding the CA-graph and to remove erroneous edges and iterative refinement
    of the clusters
    
    Parameters
    ----------
    base_clusters : list of array_like or pandas dataframe
        List of cluster labels where each set of labels must be in the same order
        and have shape (n_objects,). Note: standard output from sci-kit learn 
        clustering algorithms will return this format. If pandas dataframe, 
        should be shape (n_objects, n_clusters)
    metaclustering_alg : string
        string to denote how to cluster the object-to-object cluster association
        graph. Options are: 'louvain','spectral','METIS'. Default is 'louvain'
    n_clusters : int
        Number of clusters to find. Used only for spectral and METIS. Default
        is None
    threshold : float
        The threshold for locally pruning the lowest ratio of entries
        in the cluster association matrix. Default is 0.3
    cluster_refinemnt : boolean
        Wether to use iterative refinement of the clusters in the ensembling 
        procedure. Default is True
    max_refinement_steps : int
        The max number of refinement iterations to run, if doing iterative
        refinement. Default is 10
    refinement_number : int
        The size of the ensemble to use in each refinement step. Default is 10
    
    Returns
    -------
    z : ndarray
        cluster labels of shape (n,).
    """
    
    def __init__(self, metaclustering_alg = 'louvain', n_clusters = None,
                 threshold =0.3, cluster_refinement=True, refinement_number = 10, 
                 tolerance=0.1, max_refinement_steps=10):
        
        self.metaclustering_alg = metaclustering_alg
        self.n_clusters = n_clusters
        self.tau = threshold
        self.cluster_refinement = cluster_refinement
        self.refinement_number = refinement_number
        self.tol = tolerance
        self.max_iter = max_refinement_steps
    
    
    def fit_predict(self, base_clusters):
        BA = self._produce_ba_matrix(base_clusters)
        clusterer = graph_clustering_alg()
        CA = 1/(len(base_clusters)) * BA @ BA.T
        CA = self._prune(CA)
        self.ca_graph = CA.copy()
        self.refinement_iter =1
            
        if self.cluster_refinement:
            converged = False
            new_clusters = [clusterer.cluster(CA, self.n_clusters, self.metaclustering_alg) for _ in range(self.refinement_number)]
        else:
            converged = True
            new_clusters = clusterer.cluster(CA, self.n_clusters, self.metaclustering_alg)
            
        while not converged:
            BA = self._produce_ba_matrix(new_clusters)
            CA = 1/(self.refinement_number) * BA @ BA.T
            if (np.all(CA[np.nonzero(CA)] >= (1-self.tol))) or (self.refinement_iter > self.max_iter):
                converged = True
                new_clusters = clusterer.cluster(CA, self.n_clusters, self.metaclustering_alg)
            elif self.refinement_iter >= self.max_iter:
                converged = True
                new_clusters = clusterer.cluster(CA, self.n_clusters, self.metaclustering_alg)
            else:
                CA = self._prune(CA)
                self.ca_graph = CA.copy()
                new_clusters = [clusterer.cluster(CA, self.n_clusters, self.metaclustering_alg) for _ in range(self.refinement_number)]
                self.refinement_iter += 1
                
        return new_clusters
        
        
    def _produce_ba_matrix(self, clusters):
        '''
        Internal function for converting list of clustering labels into a
        binary-association matrix that is (object,clusters)
        '''
        
        if isinstance(clusters, pd.DataFrame):
            clusters = [clusters.loc[:,i].to_numpy(dtype=int) for i in clusters.columns]
        else:
            clusters = [clustering.astype(np.int64) for clustering in clusters]
            
        ba_matrices =[]
        for cluster in clusters:
            ba_matrix = np.zeros((cluster.size, cluster.max()+1))
            ba_matrix[np.arange(cluster.size), cluster] = 1
            ba_matrices.append(ba_matrix)
        return np.concatenate(ba_matrices, axis=1)
    
    def _prune(self, adjacency):
        '''
        Internal function for doing a global-local thresholding of an
        adjacency matrix of a graph
        '''
        thresh = np.apply_along_axis(lambda row: np.quantile(row[np.nonzero(row)], self.tau), 1, adjacency)
        mins = np.apply_along_axis(lambda row: np.min(row[np.nonzero(row)]), 1, adjacency)
        global_thresh = np.quantile(adjacency[np.nonzero(adjacency)], self.tau)
        thresh[mins > global_thresh] = 0
        adjacency[adjacency < thresh] = 0
        # adjacency[adjacency < global_thresh] = 0
        return adjacency
        
    
    
class graph_clustering_alg:
    '''
    Factory class that does the clustering of the CA graphs for the other
    method classes
    '''
    def cluster(self, graph, n_clusters, format):
        clustering = self._get_clustering(format)
        return clustering(graph, n_clusters)
    
    def _get_clustering(self, format):
        if format == 'louvain':
            return self._graph_louvain
        elif format == 'spectral':
            return self._spectral
        elif format == 'METIS':
            return self._metis
        else:
            raise ValueError(format)
            
    def _spectral(self, graph, n_clusters):
        clstr = SpectralClustering(n_clusters = n_clusters, affinity='precomputed', n_jobs=-1)
        return clstr.fit_predict(graph)
    
    def _graph_louvain(self, graph, n_clusters):
        cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=10)]
        parts = [clstr.fit(graph).labels_ for clstr in cluster_iterations]
        modularities = [modularity(graph, part) for part in parts]
        return parts[np.argmax(modularities)]
    
    def _metis(self, ca_adjacency, n_clusters):
        try:
            import nxmetis
        except:
            print("METIS is not able to import.")
        binary_adjacency = ca_adjacency.copy()
        binary_adjacency[binary_adjacency>0]=1
        ca_graph = nx.from_numpy_array(binary_adjacency)
        _, temp_clusters = nxmetis.partition(ca_graph, n_clusters)
        meta_clusters = np.zeros(ca_adjacency.shape[0])
        for i in range(len(temp_clusters)):
            for j in range(len(temp_clusters[i])):
                meta_clusters[temp_clusters[i][j]] = i
        return meta_clusters+1
        
        
        
        
        
        
        
        